//GC9A01 Display and NTP Server
#include <WiFi.h>
#include "time.h"
#include <Adafruit_GFX.h>
#include <Adafruit_GC9A01A.h>
#include <SPI.h>

#include <ArduinoBLE.h>
// See the following for generating UUIDs:
// https://www.uuidgenerator.net/

BLEService PlantyService("34bb9231-b661-49ed-9c9e-94664503c001");  // Service UUID
BLEStringCharacteristic measureCharacteristic("1d20be8f-b719-4c40-aa2a-985d72ebaff9", BLERead | BLEWrite | BLENotify, 20);
BLEIntCharacteristic thresholdCharacteristic("3c106432-ed69-4c13-a0ae-7bf960b90754", BLERead | BLEWrite | BLENotify);
BLEIntCharacteristic WaterAvgCharacteristic("a50a6058-9c22-42bf-ad25-95e3b784ddc7", BLERead | BLEWrite | BLENotify);

// Replace with your Wi-Fi credentials
const char* ssid = "San3a Tech Network";
const char* password = "WeLoveSan3aNetwork";

// NTP Server
const char* ntpServer = "pool.ntp.org";
const long gmtOffset_sec = 7200;      // Egypt TimeZone UTC+2
const int daylightOffset_sec = 3600;  // Daylight Time UTC+1

// Time tracking
int lastLoggedHour = -1;  // initialize to invalid hour

#define TFT_DC D7
#define TFT_CS D6
#define TFT_MOSI D10
#define TFT_SCK D8

Adafruit_GC9A01A tft(TFT_CS, TFT_DC, TFT_MOSI, TFT_SCK, -1);
//------------------------------------------------------------
//Neopixel Stick
#include <Adafruit_NeoPixel.h>
#define PIN A0
#define NUMPIXELS 16
Adafruit_NeoPixel pixels(NUMPIXELS, PIN, NEO_GRB + NEO_KHZ800);
//------------------------------------------------------------
//Pump
int Pump = 9;

bool done = false;
//------------------------------------------------------------
//Water Level Sensing
long result;  //variable for the result of the tx_rx measurement.
int analog_pin = A2;
int tx_pin = D3;
//------------------------------------------------------------
//Moisture Sensor
int moistRead;
int sensor_pin = D1;
//------------------------------------------------------------

//Water Level
long tx_rx() {  //Function to execute rx_tx algorithm and return a value
                //that depends on coupling of two electrodes.
                //Value returned is a long integer.
  int read_high;
  int read_low;
  int diff;
  long int sum;
  int N_samples = 100;  //Number of samples to take.  Larger number slows it down, but reduces scatter.

  sum = 0;

  for (int i = 0; i < N_samples; i++) {
    digitalWrite(tx_pin, HIGH);          //Step the voltage high on conductor 1.
    read_high = analogRead(analog_pin);  //Measure response of conductor 2.
    delayMicroseconds(100);              //Delay to reach steady state.
    digitalWrite(tx_pin, LOW);           //Step the voltage to zero on conductor 1.
    read_low = analogRead(analog_pin);   //Measure response of conductor 2.
    diff = read_high - read_low;         //desired answer is the difference between high and low.
    sum += diff;                         //Sums up N_samples of these measurements.
  }
  return sum;
}  //End of tx_rx function.


//Pump
void pump() {
  digitalWrite(Pump, HIGH);  // Control speed
  for (int i = 0; i < NUMPIXELS; i++) {
    pixels.setPixelColor(i, pixels.Color(0, 0, 150));  //Blue
  }
  pixels.show();
  delay(1000);

  digitalWrite(Pump, LOW);
  done = true;
}


//Moisture Reading
void ReadMoistSensorOne() {
  int moistRead = analogRead(sensor_pin);
  moistRead = map(moistRead, 4095, 1700, 0, 100);
  
  int threshold = thresholdCharacteristic.value();
  Serial.println(moistRead);
  tft.fillRect(0, 90, 240, 30, GC9A01A_BLACK);
  tft.setCursor(105, 90);
  tft.setTextColor(GC9A01A_YELLOW);
  tft.setTextSize(2);
  tft.print(moistRead, 1);
  tft.println("%");


  if (moistRead > threshold) {
    for (int i = 0; i < NUMPIXELS; i++) {
      pixels.setPixelColor(i, pixels.Color(0, 150, 0));  //Green
    }
    pixels.show();
    tft.fillRect(70, 120, 240, 30, GC9A01A_BLACK);
    tft.setCursor(95, 120);
    tft.setTextColor(GC9A01A_GREEN);
    tft.println("Happy");
    delay(5000);
    tx_rx();
    result = tx_rx();
    result = map(result, 301000, 225000, 100, 0);
    int WaterAvg = WaterAvgCharacteristic.value();
    tft.fillRect(50, 180, 240, 240, GC9A01A_BLACK);
    tft.setCursor(50, 180);
    tft.setTextColor(GC9A01A_BLUE);
    tft.println("Water Level:");
    tft.setCursor(105, 205);
    tft.setTextColor(GC9A01A_BLUE);
    tft.println(result);
    if (result > WaterAvg) {
      for (int i = 0; i < NUMPIXELS; i++) {
        pixels.setPixelColor(i, pixels.Color(0, 150, 0));  //Green
      }
      pixels.show();
    } else {
      for (int i = 0; i < NUMPIXELS; i++) {
        pixels.setPixelColor(i, pixels.Color(150, 0, 0));
      }
      pixels.show();
    }
  } else {
    for (int i = 0; i < NUMPIXELS; i++) {
      pixels.setPixelColor(i, pixels.Color(150, 0, 0));
    }
    pixels.show();
    tft.fillRect(70, 120, 240, 30, GC9A01A_BLACK);
    tft.setCursor(95, 120);
    tft.setTextColor(GC9A01A_RED);
    tft.println("SAD");
    delay(5000);
    tx_rx();
    result = tx_rx();
    result = map(result, 301000, 225000, 100, 0);
    int WaterAvg = WaterAvgCharacteristic.value();
    tft.fillRect(50, 180, 240, 240, GC9A01A_BLACK);
    tft.setCursor(50, 180);
    tft.setTextColor(GC9A01A_BLUE);
    tft.println("Water Level:");
    tft.setCursor(105, 205);
    tft.setTextColor(GC9A01A_BLUE);
    tft.println(result);
    if (result > WaterAvg) {
      pump();
      for (int i = 0; i < NUMPIXELS; i++) {
        pixels.setPixelColor(i, pixels.Color(0, 150, 0));  //Green
      }
      pixels.show();
    } else {
      for (int i = 0; i < NUMPIXELS; i++) {
        pixels.setPixelColor(i, pixels.Color(150, 0, 0));
      }
      pixels.show();
    }
  }
}

String oldH, oldM, oldS;

//NTP Server and Displaying Time
void displayTime(struct tm timeinfo) {
  char buf[16];
  strftime(buf, sizeof(buf), "%H:%M:%S", &timeinfo);
  String formattedTime = String(buf);
  String newH = formattedTime.substring(0, 3);  // "13:"
  String newM = formattedTime.substring(3, 6);  // "45:"
  String newS = formattedTime.substring(6, 8);  // "30"

  //tft.fillRect(20, 150, 240, 20, GC9A01A_BLACK);  // Clear only time area
  if (newH != oldH) {
    tft.fillRect(107, 150, 24, 16, GC9A01A_BLACK);
  }
  if (newM != oldM) {
    tft.fillRect(143, 150, 24, 16, GC9A01A_BLACK);
  }
  if (newS != oldS) {
    tft.fillRect(179, 150, 24, 16, GC9A01A_BLACK);
  }

  tft.setCursor(35, 150);
  tft.setTextColor(GC9A01A_CYAN);
  tft.setTextSize(2);
  tft.print("Time: ");
  tft.print(newH);
  tft.print(newM);
  tft.print(newS);
  //tft.print(buf);

  oldH = newH;
  oldM = newM;
  oldS = newS;
}

void setup() {
  Serial.begin(115200);
  pinMode(Pump, OUTPUT);
  digitalWrite(Pump, LOW);
  pixels.begin();
  pixels.show();
  // Start display
  tft.begin();
  tft.fillScreen(GC9A01A_BLACK);
  tft.setRotation(2);

  // Connect to Wi-Fi
  WiFi.begin(ssid, password);
  Serial.print("Connecting to WiFi");
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }
  Serial.println("Connected!");

  // Init time via NTP
  configTime(gmtOffset_sec, daylightOffset_sec, ntpServer);

  // Wait until time is synced
  struct tm timeinfo;
  while (!getLocalTime(&timeinfo)) {
    Serial.println("Waiting for time...");
    delay(1000);
  }

  tft.setCursor(65, 20);
  tft.setTextColor(GC9A01A_GREEN);
  tft.setTextSize(3);
  tft.println("Planty");

  tft.setCursor(65, 60);
  tft.setTextColor(GC9A01A_YELLOW);
  tft.setTextSize(2);
  tft.println("Moisture:");


  pinMode(tx_pin, OUTPUT);

  ReadMoistSensorOne();  // Take an immediate first reading
  lastLoggedHour = timeinfo.tm_hour;

  // BLE
  thresholdCharacteristic.writeValue(40);
  WaterAvgCharacteristic.writeValue(40);
  if (!BLE.begin()) {  // begin initialization
    while (1)
      ;  // wait until initialization complete
  }

  BLE.setLocalName("BLE_Planty");                           // set advertised local name
  BLE.setAdvertisedService(PlantyService);                  // set advertised service UUID
  PlantyService.addCharacteristic(measureCharacteristic);   // add the red characteristic to the service
  PlantyService.addCharacteristic(thresholdCharacteristic);     // add the green characteristic to the service
  PlantyService.addCharacteristic(WaterAvgCharacteristic);
  BLE.addService(PlantyService);                            // add service
  BLE.advertise();                                          // start advertising
}

void loop() {
  unsigned long tick = millis();
  digitalWrite(Pump, LOW);
  struct tm timeinfo;
  if (getLocalTime(&timeinfo)) {
    int currentHour = timeinfo.tm_hour;
    displayTime(timeinfo);
    // Check if current hour is divisible by 3 and different from last logged hour
    if ((currentHour % 3 == 0) && (currentHour != lastLoggedHour)) {
      ReadMoistSensorOne();
      lastLoggedHour = currentHour;
    }
  } else {
    Serial.println("Failed to get time.");
  }
  unsigned long tock = millis();

  unsigned long delayInterval = 1000 - tock + tick;
  // int moistRead = analogRead(sensor_pin);
  // moistRead = map(moistRead, 4095, 0, 0, 100);
  // Serial.println(moistRead);
  delay(int(delayInterval));
  Serial.println("No Bluetooth Connection");
  
  

  BLEDevice central = BLE.central();  // listen for BLE devices to connect:

  if (central) {  // if a central is connected to peripheral:
    // put code here to preform 1 time when device is connected
    Serial.println("Connected");
    while (central.connected()) {  // while the central is still connected to peripheral
                                   // if there is an update from Android App, change light ON (0) or OFF (1)
      if (measureCharacteristic.written()) {
        String value = measureCharacteristic.value();
        Serial.println("Recieved: " + value);
        if (value == "0") {
          
          int moistRead = analogRead(sensor_pin);
          moistRead = map(moistRead, 4095, 1700, 0, 100);
          String moistString = "M" + String(moistRead);
          measureCharacteristic.writeValue(moistString.c_str());
          Serial.println("Sent: " + moistString);
        } else if (value == "1") {
          // Water level reading
          long result = tx_rx();
          result = map(result, 301000, 225000, 100, 0);
          String resultString = "T" + String(result);
          measureCharacteristic.writeValue(resultString.c_str());
          Serial.println("Sent: " + resultString);
        } else if (value == "2") {
          // Water level reading
          ReadMoistSensorOne();
          if (done) {
            String pump1String = "P1";
            measureCharacteristic.writeValue(pump1String.c_str());
            Serial.println("Sent: " + pump1String);
            done = false;
          } else {
            String pump2String = "P2";
            measureCharacteristic.writeValue(pump2String.c_str());
            Serial.println("Sent: " + pump2String);
          }
        }
      }
    }
    Serial.println(".");
  }
}

